/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    source code: ECMAScript 2009 - ES5  (ECMAScript 5, JavaScript 5)
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

//  xxxxxxxxxxxxxxx           version: 2.5           xxxxxxxxxxxxxxxx 


// variable research parameters  => these variables are directly changed by the user within the Options Form interface
var lc = 4;    // minimum (low) count visible shapes
var hc = 8;   // maximum (high) count visible shapes. 
var w = 1;    // number of item within each level between lc..hc
                                // see bellow clock= basic timer resolution in ms  (timer period) =100ms = const
var wait = 6;           // number of basic time cycles before main stimulus (geometric shape) exposition. Units = number of basic timer resolution (clock=100ms). Time of displaying the fixing cross
var time = 1;           // stimulus exposition time in ms. Units = number of basic timer resolution (clock=100ms)
var delay = 10;       // delay between stimulus and user response. Units = number of basic timer resolution (clock=100ms)
  //var shape = 1;      // 1 = Circle, 3 = Triangle, 4 = Square/Rhomboid, 5 = Polygon-5 sides, 6 = Polygon-6 sides   
var posit = 1;          //  1= full screen;   2 = left-full-right;  3 = quadrants (up-right, down-right,left-down, left-up)  
var monochrome= true;   // if true monochrome (black) color circle.  false = circles  will be in random Colors 
var isCross = true;       // if isCross=true =>  The cross will appear in the center of the screen before the main stimulus
	// var demo = false;
	// var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser.  


// global fixed parameters => these variables were not directly available to the user
var clock = 100;      // basic timer resolution in ms  (timer period) => user can't change clock (hardcoded) = const
var itemCnt;    // total item count => itemCnt = w*(hc-lc+1);     
var timerID = null;   // timerID is the identifier of the repeated action you want to cancel. This ID is returned from setInterval(). 
var item = 0;   // current item
var step = -1;   // global timer counter 
var cor = [];   // correct number of visible shapes for each item
var ans = [];  //  user answers for each item (dependent var)
var stim= [];   // for each item contains visible cells. jagged array [item][cells] => within each [item] is separate array with ordinal numbers of visible cells  
var pos = [];     // screen position for each item (used only if window posit !== 1)
var err;          // number of Print Screen keyboard pressed during task 
var dirty = false;  // if  (dirty==true) => user changed any options in form
var is_storage= false; // local storage is enabled on user browser or not
var startTime;    // start time for task
var raster;       // screen/viewport raster (cell raster)
var is_task = false;    // task in progress or not. is_task == true for all events between startTask() and theEnd() events


// init task
function init() 
{    
  if (timerID)  {   
    window.clearInterval(timerID);  
  }  
  timerID= setInterval(function () {appTimer()},clock); // Returns an intervalID. intervalID is a unique interval ID you can pass to clearInterval()  
  
	is_storage= isStorage() && storageAvailable('localStorage');
  if (is_storage)  {
    document.getElementById('store_group').style.display= 'block';
		document.getElementById('no_storage').style.display= 'none';
  }	
	
	if (!supportedTransform())  {
    doCenter(document.getElementById('start'));
    doCenter(document.getElementById('response'));
  }    
  document.onkeyup = catchPS;    
  
	var e= document.getElementById('answer');  // attach event handlers..
	e.onkeypress = answerKeyPress;	
	e.onkeyup = answerKeyUp;	 	
	e.onpaste= function(){return false}; // Prevent pasting in answer input type=text (since pasted content might include non-allowed characters)
	    //e.onpaste = event => false; // work only in ECMAScript 2015 (not in IE11)
	
	setIns();
	showStart(); 
}


// application timer handler function
// basic timer resolution is aprox 100 ms == clock
function appTimer()  
{             
   if (step >= 0)   // else nop, idle, do none, Timer no operation
	 { 	 
		 step++; 		 
		 if (step < wait - 1 && isCross)  
		 {   // time before main stimulus (points) period
				if (step % 2 != 0)  {     // always trigger this in first timer call
					drawCross();     // draw Cross only if it is odd number (1,3,5,..)				
				}
				else   {       // for even numbers  (2,4,6,..) => clear previous cross
					clear();     // clear canvas => clear Cross in 2,4,6 step   				
				}        
		 }   
		 else if ( step == wait-1 && isCross )  {   // ALWAYS clear canvas before main stimulus display
			 clear();
		 }			
		 else if (step == wait) {
			 drawStim();     // draw main stimulus
		 }
		 else if (step == wait+time) {
			 clear();      // clear main stimulus 
		 }
		 else if (step >= wait + time + delay) {
		   step= -1;         // nop, idle, do none, Timer no operation
		   showResponse();       				
		 }
	 }
}


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== 'undefined') {
     return true;                 
  } 
  else {
    return false;    
  }  
}


// necessarily because of IE11 offline mode
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


//  check for browser supported transform
function supportedTransform() 
{
  var prefixes = 'transform WebkitTransform MozTransform OTransform msTransform'.split(' ');
  var div = document.getElementById('start');
  for(var i = 0; i < prefixes.length; i++) {
    if(div && div.style[prefixes[i]] !== undefined) {
      return prefixes[i];
    }
  }
  return false;
}


// if no css transform do absolute position for html elements (divs). center div(s)
function doCenter(e)  
{
  if (e.id === 'start')  {        
    e.style= (window.innerWidth > 600 && window.innerHeight > 600 ) 
    ? 'position:absolute;min-width:50%;min-height:50%;margin:auto;overflow:auto;'  
    : 'position:absolute;min-width:75%;min-height:75%;margin:auto;overflow:auto;';  
  }
  else if (e.id === 'response')  {
    e.style='position:absolute;min-width:25%;min-height:25%;margin:auto;overflow:auto;';    
  } 
}


// deprecated but still works: https://developer.mozilla.org/en-US/docs/Web/API/KeyboardEvent/keyCode
// https://developer.mozilla.org/en-US/docs/Web/API/KeyboardEvent/code
// https://caniuse.com/#search=KeyboardEvent.keyCode
// https://caniuse.com/#search=KeyboardEvent.code
// This function is implemented because of possible cheating (we can not prevent the use of cameras on mobile phones)
function catchPS(e)   
{ 
  e = e || window.e;  
  if (e.key === 'PrintScreen' && step >= 0) {  
    err++;  
  } 
}


// raised when page is resized. see html: <body onresize="pageResize()"
function pageResize()  
{ 
	if (is_task) 	{ 
    var c= document.getElementById('can');
		c.width= window.innerWidth;  
		c.style.width= c.width + 'px';	     
		c.height= window.innerHeight;   
		c.style.height= c.height + 'px';  // https://css-tricks.com/whats-the-difference-between-width-height-in-css-and-width-height-html-attributes/ 		 						
	}
}	
 
 
// show start (instruction) panel
function showStart()  
{      
  document.getElementById('end').style.display= 'none';
  document.getElementById('google-wrapper').style.display='none';  
  document.getElementById('start').style.display= 'block';
	document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
}


// start task 
function startTask()   
{      
  if ( (posit == 2 && w %  3 != 0) || (posit == 3 && w % 4  != 0) )  
	{                                                                                                   
    var s= 'Invalid Number of tasks within each level for selected Position of shapes on the screen. Number of tasks within each level must be ';
    s= (posit === 2 )  ? s + 'number must be divisible by 3 without a remainder (3,6,9,..)' : s + 'number must be divisible by 4 without a remainder (4,8,12,..)';
    document.getElementById('is_valid').textContent= s;    
    document.getElementById('is_valid').setAttribute('class', 'error');    
    return;                                                                                         
  } 
   
	initLayout();
  setPar();
  setRaster();
	stimGen();
  stim= shuffle(stim);    
  computeCor(); 
              
  startTime = new Date();						
  item= 0;
  is_task= true; 
	pageResize();  // after is_task= true => set canvas
	step=0;    // "activate" Timer       
	
}


// init visible/invisible html dom elements. init canvas 
function initLayout()  
{
	document.getElementById('google-wrapper').style.display='none';
  document.getElementById('end').style.display= 'none'; 
  document.getElementById('start').style.display= 'none'; 
	document.getElementById('demo_sol').style.display= document.getElementById('demo').checked ? 'block' : 'none';
	document.getElementById('can').style.display= 'block';	// **
}


// set task parameters (they are not available to the user directly)
function setPar()  
{   
  itemCnt = w * (hc-lc+1);    
	stim = [];  // jagged stimulus array     
  cor = [];   // correct number of visible shapes within each item
  ans = [];   // type == string (for google charts) user answers (dependent var)   
	pos= [];    // screen position for each item (used only if window posit !== 1)
  err= 0;  
}


// set value of hidden screen cell raster
function setRaster()  
{
	raster= Math.ceil(Math.sqrt(hc)); 
	if ( posit == 2 || posit == 3)  {
		raster = raster * 2 + 1;
  }
	if (raster % 2 == 0)  {
		raster++;  // muset be odd number
	}	
}


// stimulus generator => generate visible cells ih hidden rectangle grid
function stimGen()  
{                   
  var rst= raster, tmp;  
	var cells= lc;   // number of visible cells (shapes). range= [lc..hc]
  for (var i=0; i<itemCnt; i++)  // i == ordinal item number
	{      
    stim[i]= [];
		
		for (var j= 1; j <= cells; j++)  // shape counter
		{			
			
			if (posit == 1)  {       
        do  {
					tmp= getCell(0, rst);  // full screen
        }	while ( exists( tmp, stim[i] ) );
				stim[i].push([tmp]);				
      }  
      
			else if (posit == 2)  
			{ //  left - full screen - right => items will be mixed (shuffle) later; from lf to hc (w-times)  MUST be: (itemCnt %  3 === 0 && w % 3 ===0)
        switch ( i % 3 )  
				{
          case 0:					
						do {			
							tmp= getCell(1, rst);  // 1 => left half of screen	
						} while ( exists( tmp, stim[i] ) );
					  stim[i].push([tmp]);			
						pos[i]= 1; // left part of screen
						break;
					
          case 1:  				  
						do  {
							tmp= getCell(0, rst);  // full screen
						}	while ( exists( tmp, stim[i] ) );
						stim[i].push([tmp]);		    
						pos[i]= 2; // full screen
						break;
					
          case 2:  // or default:				            
						do {			
							tmp= getCell(2, rst);  // 2 => right half of screen 
						} while ( exists( tmp, stim[i] ) );
						stim[i].push([tmp]);		          					
						pos[i]=3;  // right part of screen  
						break;					
        }  // switch                                                                     
      }      
      
			else if (posit == 3)    
			{ //  4 quadrants: up-left, up-right, down-right, down-left  => items will be mixed (shuffle) later => from lf to hc (w-times) =>  MUST be: (itemCnt %  4 === 0 && w % 4 ===0)            
        switch ( i % 4 )  
				{
          case 0: 
						do {			
							tmp= getCell(3, rst);  // 3 => top-left quadrant (q1)	
						} while ( exists( tmp, stim[i] ) );
						stim[i].push([tmp]);		            
						pos[i]= 1;  // up-left 
            break;
						
          case 1:   
						do {			
							tmp= getCell(4, rst);  // 4 => top-right quadrant (q2)	
						} while ( exists( tmp, stim[i] ) );            
						stim[i].push([tmp]);		
						pos[i]= 2;  // up-right
            break;
						
          case 2:   
						do {			
							tmp= getCell(5, rst);  // 5 =>  down-right quadrant (q3) 
						} while ( exists( tmp, stim[i] ) ); 
						stim[i].push([tmp]);	
						pos[i]= 3; // down-right   
            break; 
						
          case 3:     // default:        
            do {			
							tmp= getCell(6, rst);  // 6 => down-left quadrant (q4) 
						} while ( exists( tmp, stim[i] ) );
						stim[i].push([tmp]);		
						pos[i]= 4; // down-left 
            break; 
        }            
      } 
		}  // for (var j= 1; j <= cells; j++)  // shape counter
		
		if ( (i + 1) % w == 0 ) {
			cells++;   // increase shape number
		}
		
	}  //for (var i=0; i<itemCnt; i++)			
  
}


/* spatial => [0 = full screen; 1 => left half, 2 = right half, 3 = top-left quadrant; 4 = top-right; 5 = down-right, 6 = down-left quadrant] 
   rst = window.raster => canvas cells raster
	 return position of target cell in raster  */
function getCell(spatial, rst)
{    
	var N = rst * rst - 1;   // N = maximum number of cells in canvas object
	var pos, row, col, not_ok; // pos => return value => index (position) of target cell 	
			// row = Math.floor (pos / rst);  col = Math.floor (pos % rst); 
			// or => col= pos - row * rst;  // pos = row * rst + col; 				
	var half= Math.floor(rst/2);
	
	if ( spatial == 0 )  {	  // full screen (0)
		pos= getRndInt(0,N);    // random selected cell position
	}
  else  
	{	
		do		
		{
			pos= getRndInt(0,N);  // random selected cell position
			row = Math.floor (pos / rst);  // random selected row
			col = pos % rst;     // random selected column
			// set do-while loop iteration condition
			switch (spatial)
			{				
				case 1:  // left half of screen	 			
					not_ok = ( col >= half );  
					break;

				case 2:  // right half of screen 
					not_ok = ( col <= half );
					break;		

				case 3:  // top-left quadrant (q1)
					not_ok= ( row >= half || col >= half );	
					break;		
				
				case 4:  // top-right quadrant (q2) 
					not_ok= ( row >= half ||  col <= half );	
					break;		
						
				case 5:   // down-right quadrant (q3)
					not_ok = ( row <= half || col <= half );	
					break;		

				case 6: // down-left quadrant (q4) 
					not_ok = ( row <= half || col >= half );	
					break;

				default: // this case is error			
				  pos= getRndInt(0,N);   // random selected cell position
				}  // switch				
		}	 while (not_ok);   		
	}   // else
	return pos;	
}		


/* check if this cell position (pos) already is selected.
   if this value (cell position pos) exists return true else return false;
   pos => shape position
   arr => array to check  */
function exists(pos, arr) 
{ 
	var len= arr.length;
	for (var i= 0; i < len; i++)  {
    if (arr[i] == pos) { 
			return true;
		}
	}
  return false;	
}


// compute correct answer for each item in task
function computeCor()  
{ 
  cor= [];   
  for (var i=0; i< itemCnt; i++)  { 
    cor[i]=0;  // correct number of visible shapes within each item
    for (var j in stim[i]) {                   
      if (stim[i][j] != 'undefined' ) {
				cor[i]++; 
			}		
    }    
  }
}


// unbiased shuffle algorithm: Fisher-Yates (aka Knuth) Shuffle.
function shuffle(arr)   
{
  var cur_index = arr.length, tmp, rnd_index, tmp2;
  // While there remain elements to shuffle...
  while (0 !== cur_index)   {    
    rnd_index = Math.floor(Math.random() * cur_index);     // Pick a remaining element...
    cur_index = cur_index - 1;   
    tmp = arr[cur_index];          // And swap it with the current element.
    if (posit !== 1)  tmp2= pos[cur_index];      //  also do not forget swap screen positions
    arr[cur_index] = arr[rnd_index];
    arr[rnd_index] = tmp;     
    if (posit !== 1)  {    //  swap screen positions
      pos[cur_index] = pos[rnd_index];
      pos[rnd_index] = tmp2;       
    }
  }
  return arr;
}


// return random integer in range [min..max]
function getRndInt(min, max) {
    return Math.floor(Math.random() * (max - min + 1) ) + min;
}


// draw cross on screen center
function drawCross()  
{
  var c= document.getElementById('can'), ctx= c.getContext('2d'); 
	var x= Math.round(c.width/2), y= Math.round(c.height/2), d = Math.floor(Math.min(x,y)/4);  // d == 1/8 min(width, height)  
  drawLine(ctx, x-d, y,   x+d, y);   // horizontal
  drawLine(ctx, x,   y-d, x,   y+d); // vertical 
}


// draw line on canvas
// par: ctx => handle context; (x1, y1) => 1. point coord; (x2, y2) => 2. point coord
function drawLine(ctx, x1,y1,x2,y2)  
{
  ctx.beginPath();
  ctx.moveTo(x1,y1);
  ctx.lineTo(x2,y2);
  ctx.lineWidth= 1;
  ctx.strokeStyle='#000';  // black
  ctx.stroke();     
}


// draw stimulus (geometric shapes) on canvas 
function drawStim()  
{  
  document.body.style.cursor = 'none'; 
 	var c= document.getElementById('can'), ctx= c.getContext('2d');   
	ctx.clearRect(0, 0, c.width, c.height); // clear canvas object 		

	var rst= raster, len = stim[item].length;   // raster value, total number of shapes in this item
	var width = c.width / rst, height = c.height / rst;  //  cell width, cell height
  var rad= Math.round(0.75 * Math.min(width,height)/2); // radius for each shape in each cells (px)
  var shape= parseInt(document.getElementById('shape').value,10); // selected shape

  var m_x_off= Math.floor(width/2-rad);  // maximum x-offset within hidden rect
  var m_y_off= Math.floor(height/2-rad); // maximum y-offset within hidden rect
	var is_color= monochrome ? false: true;
	if (!is_color) {
		ctx.fillStyle= '#000'; 
	}
	
	for (var i= 0; i < len; i++)         
	{	  	
		var j= stim[item][i];   // j => cell position in hidden grid (this cell will show shape) 
		var row = Math.floor (j / rst);
		var col = j % rst;                 						
		var left= col * width, top = row * height; // cell left bound, cell top bound      
	  var x_off= getRndInt(-m_x_off,m_x_off); // random x-offset within hidden rect
    var y_off= getRndInt(-m_y_off,m_y_off); // random y-offset within hidden rect  
  	var x= left + width/2 + x_off;
		var y= top + height/2 + y_off;	
		
		ctx.beginPath();                    
		if (shape == 1) {   // draw circle      
		  ctx.arc (x, y, rad, 0, 2*Math.PI); //ctx.arc(x,y,r,sAngle,eAngle,counterclockwise); 
		}
		else  
		{  // draw regular polygon (3,4,5,6 sides)
			for (var k= 1; k<= shape; k++)  
			{            
				var px = x + rad * Math.cos(k*2*Math.PI/shape);
				var py = y + rad * Math.sin(k*2*Math.PI/shape);
				if (k == 1) {
					ctx.moveTo(px,py);
        }
        else	{
					ctx.lineTo(px,py); 
				}
			}    
		}            
    if (is_color)  {		
			ctx.fillStyle= getRandomColor();             
		}
		ctx.fill();                                  
  }  

}


// return random color in hex format
function getRandomColor() {
  var letters = '0123456789abcdef';
  var color = '#';
  for (var i = 0; i < 6; i++) {
    color += letters[Math.floor(Math.random() * 16)];     // returns a random integer from 0 to 15 
  }
  return color;
}  


// clear canvas
function clear()  {
	var c= document.getElementById('can');
  c.getContext('2d').clearRect(0, 0, c.width, c.height);     // clear canvas
}	


/* see function init().  deprecated but still works:
   https://developer.mozilla.org/en-US/docs/Web/API/GlobalEventHandlers/onkeypress
   this regex [0-9()*\/+-] was tested here:    https://regex101.com/    */
function answerKeyPress(event)   { 
  return event.charCode === 0 || /[0-9]/.test(String.fromCharCode(event.charCode));	
}


// go to the next task if the ENTER key is pressed. see function init() 
function answerKeyUp(e)  
{  
	if ( e.key === 'Enter' || e.keyCode === 13 ) {  
		nextItem();
	}
}


// validate user answer value entered in document.getElementById('answer') html element
function valAnswer()  
{	
  var v = trim(document.getElementById('answer').value);  // user answer (trimed) 
	var s = document.getElementById('answer_val');   // validator label
	if ( v == '' )  { 
	  document.getElementById('answer').value= '0';   // if there is no answer from user => set value 0
		s.textContent= ''; 		// clear if no user entry
    return true;   // user can escape answer		
	} 	
	else if ( isInt(v) && parseInt(v,10) >= 0 && parseInt(v,10) <= 99 )  {
		s.textContent= '';
		return true;
	}
	else { 
		s.textContent= 'Enter the number of between 0 and 99. If you do not know the answer: enter zero or empty (blank).';	
		return false;
	}		
}


// goto next item/task or end research
function nextItem()  
{
  var e= document.getElementById('answer');    
  if ( !valAnswer() ) {
    e.focus();   //alert('Enter the number of between 0 and 99. If you do not know the answer: enter zero or empty (blank).');
    return;
  }   
  
  var tmp = (trim(e.value) == '') ? 0 : e.value;  // redundant
  tmp = parseInt(tmp,10);         
	ans[item]= tmp;     // store validated user answer (as integer)
					
  document.getElementById('response').style.display='none';      
	
	item++;      
  if (item < itemCnt) {    // show next task    //document.getElementById('can').style.display= 'block';            
    step=0;      // activate Timer    
  }
  else  {      // end of task           
    theEnd();    
  }
}


// trim input string. str = input string
function trim(str) {
   return str.replace(/^\s\s*/, '').replace(/\s\s*$/, '');
}


// Checks whether it is an integer value or not
// isNaN("") = false: the empty string is converted to 0 which is not NaN                   
function isInt(n) {       
   return (!isNaN(n)) && (parseFloat(n) === parseInt(n, 10)) &&  (n % 1 === 0); 
} 


// show response div (panel where the user can type their answer)
function showResponse()  
{ 
  document.body.style.cursor = 'auto';  // restore cursor 
	document.getElementById('response').style.display='block';  
  if (document.getElementById('demo').checked ) {
     document.getElementById('demo_sol').textContent= 'The correct answer is ' + cor[item];
	}		
  document.getElementById('answer').value= '';
	document.getElementById('answer_val').textContent='';  // reset answer validator
	document.getElementById('answer').focus(); 
}


// return integer into string hh:mm:ss time format
// input par: total == integer (time in seconds)
function stringTime(total) 
{
  var hh = Math.floor(total/(60*60));         
  var div_min = total % (60 * 60);
  var mm = Math.floor(div_min / 60);       
  var div_sec = div_min % 60;
  var ss = Math.ceil(div_sec);    
  // These lines ensure you always have two-digits. This formats your string to HH:MM:SS
  if (hh < 10)  {hh = '0'+hh;}     
  if (mm < 10) {mm = '0'+mm;}
  if (ss < 10) {ss = '0'+ss;} 
  return hh + ':' + mm + ':' + ss;
} 


// the end of task => prepare data and calls google Chart API for report
function theEnd()  
{
	is_task= false;
	document.getElementById('end').style.display='block'; 
	document.getElementById('can').style.display= 'none';
  
	var res=0;  
  for (var j=0; j!= cor.length; j++)   {
    if (cor[j] == ans[j]) res++;
  }
  
  //document.getElementById('score').textContent='Your score is: ' + res + '   from possible: ' + cor.length + '   Percentage: ' + Math.round(100*res/cor.length) + '%';  	
	document.getElementById('score').textContent= 'Sum of points= ' + res + '. Maximum points= ' + itemCnt  + '. Percentage: ' +    Math.round(100*res/itemCnt).toFixed(3) + '%'; 
  if (err == 0)  {
		document.getElementById('p_s').style.display= 'none';
	}
	else {
		document.getElementById('p_s').style.display= 'block';
		document.getElementById('p_s').textContent= 'Number of presses of the Print Screen button during the task: ' + err;
	}
 
  document.getElementById('l_ts').textContent= 'Last completed research: ' + new Date();
  document.getElementById('l_res').style.display= 'block';
			
	document.getElementById('google-wrapper').style.display='block'; 
  document.getElementById('table1').innerHTML='';  
  document.getElementById('table2').innerHTML='';   
  document.getElementById('chart1').innerHTML=''; 
  document.getElementById('chart2').innerHTML='';   
  document.getElementById('no_google').textContent='';    
	
  var s_pos;    // s_pos => sum of points group by screen position (only if posit != 1)	
	try  
	{
    showReport1(res);
    if (posit > 1)  {
			s_pos= showReport2(res); 
		}	
  }
  catch (e)  {
    document.getElementById('no_google').textContent= 'You must have an available internet connection to display graphs and tables. Make sure you are online.'; 
  }

  try 
  {
    if (is_storage && document.getElementById('store').value === '1') {
			setStorage(res, s_pos);  
    }
		else  {
			document.getElementById('table_grp').style.display= 'none';
			document.getElementById('store_id').style.display= 'none';
		}
  }        
  catch (e)  { 
    document.getElementById('table_grp').style.display= 'block';
		document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }

}


// prepare data for google API chart/table for table1 (and chart1) 
// table1 = Displays the answer (and results) for each item.
// chart1 = Show correct answers grouped per level
function showReport1(res)   
{      
  //    xxxxxxxxxxxxxx     table1 = Displays the answer (and results) for each item        xxxxxxxxxxxxxxxxx 
  var data = new google.visualization.DataTable();    // Create and populate the data table. 
  data.addColumn('string', 'Item number');     // 1. col = ans array = must be string  
  if (posit > 1)  data.addColumn('string', 'Screen position'); 
  data.addColumn('number', 'Answer');         
  data.addColumn('number', 'Correct'); 
  data.addColumn('number', 'Item score');    
  for(var i = 0; i < ans.length; i++)  {  
    var tmp= (ans[i] === cor[i]) ? 1: 0;    
    if (posit === 1)  {
      data.addRow([(i+1).toString() + '. item',  ans[i],  cor[i], tmp]);        
    }
    else if (posit === 2)  {
       switch (pos[i])  {
         case 1:           
           data.addRow([(i+1).toString() + '. item', 'left', ans[i],  cor[i], tmp]);
           break;           
         case 2:  
           data.addRow([(i+1).toString() + '. item',  'full screen', ans[i],  cor[i], tmp]);
           break;      
         case 3:  
           data.addRow([(i+1).toString() + '. item',  'right', ans[i],  cor[i], tmp]);
       }           
    }
    else if (posit === 3)  {
      switch (pos[i])  {
        case 1:  
          data.addRow([(i+1).toString() + '. item',  'top-left', ans[i],  cor[i], tmp]);
          break;
        case 2:
          data.addRow([(i+1).toString() + '. item',  'top-right', ans[i],  cor[i], tmp]); 
          break;
        case 3:  
          data.addRow([(i+1).toString() + '. item',  'down-right', ans[i],  cor[i], tmp]);
          break;
        case  4:  
          data.addRow([(i+1).toString() + '. item',  'down-left', ans[i],  cor[i], tmp]);
      }
    }                 
  }                                   
  options= {      
    /*'width': '50%',*/    
    'allowHtml': true, 
    'sort': 'enable',  
    'alternatingRowStyle': true              
  };           
  var elem= document.getElementById('table1');
  elem.title='Total score=' + res;     // see wm4.html some title are static!!
  elem.style.display = 'block'; 
  new google.visualization.Table(elem).draw(data, options);  
  //    xxxxxxxxxxx     chart1 => BarChart => Show correct answers grouped per level  xxxxxxxxxxxxx
  var lev = {};  // correct answers for each level lc..hc  (group by level)
  for(var i= lc; i<= hc; i++)  {  
    lev[i]= 0; 
  }  
  for (var j=0; j!=cor.length; j++)   {
    if (cor[j] == ans[j]) lev[cor[j]]++;
  }    
  var data = new google.visualization.DataTable();    // create and populate new data table. 
  data.addColumn('string', 'Shapes count');     // 1. col = levels [lc..hc] = must be string             
  data.addColumn('number', 'Score within the level'); 
  data.addColumn('number', 'Maximum possible score');
  /*for (var key in lev)  {
     data.addRow([(key).toString() + '. Shapes',  lev[key],  window.w]);                                            
  } */ 
  // https://stackoverflow.com/questions/54327508/how-to-display-a-small-bar-even-when-a-value-is-zero-in-google-charts
  for (var key in lev)  {
    if (lev[key] > 0) { 
      data.addRow([(key).toString() + ' shapes',  lev[key],  window.w]); 
    }
    else  { 
      data.addRow([(key).toString() + ' shapes',   {v: 0.1, f: '0'},  window.w]);   
    }                                       
  }   
  // https://developers.google.com/chart/interactive/docs/gallery/columnchart              
  options= {
    title: 'The results are grouped by the number of displayed shapes. Total score = ' + res,
    width: '100%',  
    hAxis: {
      title: 'Number of displayed geometric shape',
    }, 
    vAxis: {
      title: 'Number of points'
    }   
  };    
  document.getElementById('chart1').style.display = 'block';   
  new google.visualization.ColumnChart(document.getElementById('chart1')).draw(data, options);     // Create and draw the visualization.      
}


// prepare data and call google API chart/table for table2 (and chart2) 
// table2 and chart2 = Displays the sum of correct answers grouped by screen positions 
// return results group by screen positions
function showReport2(res) 
{    
  var s_pos = {};  // sum of correct answers group for each screen position (1..3) or (1..4)
  var up= (posit === 2) ? 3 : 4;   // (window.posit === 2) => [0..3] else [0..4] see genStim()
  // s_pos[1] = left or up-left; s_pos[2]= full screen or up-right, s_pos[3] = right or down-right, s_pos[4]= down-left
  for(var i= 1; i<=up; i++)  {  
    s_pos[i]= 0;        
  }  
  for (var j=0; j!=cor.length; j++)   {         
    if (cor[j] == ans[j])  {
			s_pos[pos[j]]++;       
		}
  }    
  //    xxxxxxxxxxxxxx     table2 = Displays the sum of correct answers grouped by screen positions  xxxxxxxxxxxxxxxxx 
  var data = new google.visualization.DataTable();    // Create and populate the data table. 
  data.addColumn('string', 'Position');     // 1. col = ans array = must be string  
  data.addColumn('number', 'Sum of points');    // sum of correct answers grouped by screen positions      
  // we must recode pos[] and s_pos[] values based on window.posit => see genStim() 
  for(var i= 1; i<=up; i++)  
	{  
     if (posit === 2)  
		 {
       switch (i)  
			 {
         case 1:           
           data.addRow(["left", s_pos[i]]);
           break;           
         case 2:  
           data.addRow(["full screen", s_pos[i]]);
           break;      
         case 3:  
           data.addRow(["right", s_pos[i]]);
					 break;
       }           
    }
    else if (posit === 3)  
		{
      switch (i)  
			{
        case 1:  
          data.addRow(["top-left", s_pos[i]]);
          break;
        case 2:
          data.addRow(["top-right", s_pos[i]]);
          break;
        case 3: 
          data.addRow(["down-right", s_pos[i]]); 
          break;
        case 4: 
          data.addRow(["down-left", s_pos[i]]); 
					break;
      }
    }                   
  }                                    
  options= {      
    'width': '50%',    
    'allowHtml': true, 
    'sort': 'enable',  
    'alternatingRowStyle': true              
  };           
  var elem= document.getElementById('table2');
  elem.style.display = 'block'; 
  new google.visualization.Table(elem).draw(data, options); 
  //    xxxxxx      PIE chart2  => data are same   xxxxxxx
  if (res<1)  { 
    document.getElementById('chart2').innerHTML= '<p class="center">Total score = 0. The Chart will not be displayed';
    return;
  }
  options= {
   title: 'Sum of points grouped by the display position of geometric shapes on the screen',
   is3D: true,
   width: '100%'   
  };    
  document.getElementById('chart2').style.display = 'block';   
  new google.visualization.PieChart(document.getElementById('chart2')).draw(data, options);     // Create and draw the visualization.
  return s_pos;  // results group by screen positions 	
}


// refresh variable research Options (update *_msg label) => give user short option HELP
function refOpt(e)  
{   
  dirty=true;       
  var my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';    
  
	switch (my_id) 
	{
		case 'lc': 
    case 'hc':		
			if (e.validity.valid)  {           
				msg.textContent= ''; 
				//msg.removeAttribute('class');       //msg.textContent=  e.value; 
			}
			else  {
				msg.textContent= s + e.min + '..' + e.max + s2 + e.step;
				//msg.setAttribute('class', 'error');    // msg.className = 'error';
			} 
			refItem();
      break;  
  
		case 'w':      // Check interdependent with posit
			if (e.validity.valid)  {       				   
				msg.textContent= '';        //msg.textContent=  e.value; 
				//msg.removeAttribute('class');     
			}
			else  {
				msg.textContent= s + e.min + '..' + e.max + s2 + e.step;
				//msg.setAttribute('class', 'error');    // msg.className = 'error';
			}     
			var _pos= document.getElementById('posit');
			if (_pos.value === '2' && e.value %  3 != 0)  {
				msg.textContent= 'For selected Screen shape position Number of tasks within each level must be divisible by 3 with no remainder (3,6,9,..)';
				//msg.setAttribute('class', 'error');  
			}
			else if (_pos.value === '3' && e.value %  4 != 0)   {
				msg.textContent= 'For selected Screen shape position Number of tasks within each level must be divisible by 4 with no remainder (4,8,12,..)';
				//msg.setAttribute('class', 'error');  
			}           
			refItem();
			break; 
  
		case 'wait':
		case 'time':
		case 'delay':
			if (e.validity.valid)  {
				msg.textContent=  e.value + ' ms'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + '..' + e.max + s2 + e.step; 
				msg.setAttribute('class', 'error');
			}          
			break;
		
		// case 'shape': 
		
		case 'posit':    // Check interdependent with w
			var w_msg=  document.getElementById('w_msg');    // validator for window.w
			w_msg.textContent= '';
			//w_msg.removeAttribute('class');      
			if (e.value === '1')  {
				msg.textContent=  'Geometric shapes will be displayed on the full screen';      
			}             
			else if (e.value === '2')   {
				msg.textContent= '1/3 of the geometric shapes will be displayed on the left half of the screen, 1/3 on the full screen, 1/3 on the right half of the screen.';
				if (parseInt(document.getElementById('w').value, 10)  %  3 !== 0 )  {
						w_msg.textContent= 'For selected Screen shape position Number of tasks within each level must be divisible by 3 with no remainder (3,6,9,..)';
						//w_msg.setAttribute('class', 'error');
				 }
			}
			if (e.value === '3')  {
				msg.textContent= '1/4 geometric shapes will be shown top-left,1/4 of the shapes will be shown top-right, 1/4  will be shown down-left, 1/4 will be shown down-right.'; 
				if (parseInt(document.getElementById('w').value, 10)  %  4 !== 0 )  {
					w_msg.textContent= 'For selected Screen shape position Number of tasks within each level must be divisible by 3 with no remainder (4,8,12,..)';
					//w_msg.setAttribute('class', 'error');
				}
			}
			break;        

    case 'monochrome':                                                                                                                                                 
			msg.textContent= e.checked ? 'Geometric shapes will only appear in black color - monochrome' : 'Geometric shapes will be displayed in all colors (random)';
			break; 

     case 'is_cross':
		   msg.textContent= e.checked ? 'Yes. The cross will appear in the center of the screen before the main stimulus.' : 'No. There is no fixation point in the middle of the screen';     
       break;
			 
		case 'demo': 
		  msg.textContent= e.checked ? 'On. The correct solution for each item will be displayed.' : 'Off';  
		  break; 

		case 'store':                                                                                                                                            
		  msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;
	
  }  // switch 
	
  refMsg();   
}


// refresh main validator message for user defined OPTIONS (feedback to user)
function refMsg()   
{ 
 if ( optionsValid() )   {
    document.getElementById('is_valid').textContent=  (dirty===true) ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').style='text-align:center;color:#000000'; 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated. These settings are marked in red.'; 
    document.getElementById('is_valid').style= 'text-align:center;color:#ff0000;font-weight:bold'; // safari 5.1.7 bug? => this don't work
  } 
}


// update item count on options form and check lc <= hc and update  label itemCnt = w*(hc-lc+1);
function refItem()  
{
   var _lc = document.getElementById('lc');
   var _hc = document.getElementById('hc');
   var _w= document.getElementById('w');
   var _item= document.getElementById('item_cnt');
   var lc_msg= document.getElementById('lc_msg');
   var hc_msg= document.getElementById('hc_msg');    
   if (_lc.validity.valid  && _hc.validity.valid && _w.validity.valid)   
   {                      
     if  (parseInt(_lc.value,10) > parseInt(_hc.value,10))  
     {
       lc_msg.textContent= 'The starting level must not be higher than the final level';  
       hc_msg.textContent= 'The starting level must not be higher than the final level';
       //lc_msg.setAttribute('class', 'error');
       //hc_msg.setAttribute('class', 'error');    
       _item.textContent= 'Total number of items: Invalid value'; 
       _item.setAttribute('class', 'error');
     } 
     else   {                                                                                               
       //lc_msg.removeAttribute('class');    //lc_msg.textContent= _lc.value;
       //hc_msg.removeAttribute('class');   //hc_msg.textContent= _hc.value; 
       lc_msg.textContent= '';
			 hc_msg.textContent= '';
			 _item.textContent=  'Total number of items: ' +  ( parseInt(_w.value,10) * (parseInt(_hc.value,10)-parseInt(_lc.value,10)+1));      
       _item.removeAttribute('class');
     }
   } 
   else  
   {
     _item.textContent= 'Total number of items: Invalid value'; 
     _item.setAttribute('class', 'error');
   }    
}


// set user variable research parameters  => these variables are directly changed by the user within the Options Form interface
function setVarPar()  
{   
  if (document.getElementById('lc').validity.valid && document.getElementById('hc').validity.valid 
    && (parseInt(document.getElementById('lc').value,10) <= parseInt(document.getElementById('hc').value, 10) ) )   {  
      lc= parseInt(document.getElementById('lc').value, 10); 
      hc= parseInt(document.getElementById('hc').value, 10); 
  }
  if (document.getElementById('w').validity.valid) {
		w= parseInt(document.getElementById('w').value, 10);        
  }
	if (document.getElementById('wait').validity.valid) {
		wait= Math.round(parseInt(document.getElementById('wait').value, 10)/clock);
  }
	if (document.getElementById('time').validity.valid) {
		time= Math.round(parseInt(document.getElementById('time').value, 10)/clock);  
  }
	if (document.getElementById('delay').validity.valid)  {
		delay= Math.round(parseInt(document.getElementById('delay').value, 10)/clock);     
  }
	  
	posit=  parseInt(document.getElementById('posit').value, 10);     // shapes positions on screen
  var e=   document.getElementById('posit');
  var s = 'Warning:  Screen shape position: ' + e.options[e.selectedIndex].textContent;
  if (posit == 2 && w %  3 != 0)   {     
    w= 3;             // Automatic set to default values
    document.getElementById('w').value= w;   // wm4.html refresh/update
    s= s + ' must have Number of tasks within each level must be divisible by 3 with no remainder (3,6,9,..). Set to default value 3. This change also includes changing the number of items in the task';
    document.getElementById('is_valid').textContent= s;       
    document.getElementById('is_valid').setAttribute('class', 'error');                                                                                      
  }
  else if  (posit == 3 && w % 4 != 0)   {   
    w= 4;      // Automatic set to default values
    document.getElementById('w').value= w;   // wm4.html refresh/update
    s = s +  ' must have Number of tasks within each level must be divisible by 4 with no remainder (4,8,12,..). Set to default value 4. This change also includes changing the number of items in the task';    
    document.getElementById('is_valid').textContent= s;      
    document.getElementById('is_valid').setAttribute('class', 'error');                                                                                     
  } 
	
	monochrome= document.getElementById('monochrome').checked ? true: false;  // if true monochrome (only black) color circle.  false = circles  will be in random Colors  
	
  isCross= document.getElementById('is_cross').checked ? true: false;   // if (isCross==true) => The cross will appear in the center of the screen before the main stimulus        
 
}


// setVarPar() is called before this function  => set default w values  if (!w_ok) => w=3 || w=4
function optionsValid()  
{
  var range_ok= parseInt(document.getElementById('lc').value,10) <= parseInt(document.getElementById('hc').value,10);    
  var w_ok = !((document.getElementById('posit').value === '2' && parseInt(document.getElementById('w').value, 10) %  3 !== 0) || ( document.getElementById('posit').value === '3' && parseInt(document.getElementById('w').value,10) % 4  !== 0));  
  var res=   document.getElementById('lc').validity.valid &&  document.getElementById('hc').validity.valid && 
             document.getElementById('w').validity.valid  &&  document.getElementById('wait').validity.valid && 
             document.getElementById('time').validity.valid &&  document.getElementById('delay').validity.valid &&
             range_ok  &&  w_ok;
  return res;		
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById('hide_option').style.display ='block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('child');  // don't work in <= IE 9    
  document.getElementById('start_div').style.display='none';   
  document.getElementById('show_option').style.display ='none';  
  document.getElementById('hide_option').style='text-align:center';  // bug? auto-remove style when element.style.display='none'
	document.getElementById('instructions').style.display= 'none'; 
	document.getElementById('store_oper').style.display= 'none';
	if (is_storage && localStorage.getItem('wm4-v2.5')) 
  {
    var cnt= JSON.parse(localStorage.getItem('wm4-v2.5')).length;                                                        
    document.getElementById('store_cnt').textContent=  'You have a total of ' + cnt  + ' results stored in your web browser.';
    if (cnt > 0) {
			document.getElementById('store_oper').style.display= 'table';   // xxx
		}
	}
  dirty= false;
  document.getElementById('is_valid').textContent='';  // tmp reset validator message
	if ( document.getElementById('l_res').style.display == 'block' ) { 
		document.getElementById('last_data_sel').options[1].disabled= posit == 1 ? true: false;
	}
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user defined research Options
function hideOption()  
{
  setVarPar();   
  document.getElementById('options').style.display= 'none'; 
  document.getElementById('start').classList.add('child');   
  document.getElementById('show_option').style.display= 'block';   
  document.getElementById('start_div').style.display= 'block';     
  document.getElementById('hide_option').style.display= 'none';   
  document.getElementById('instructions').style.display= 'block';
	//document.getElementById('start_div').style= 'text-align:center';  // bug?  (firefox auto remove class 'center' when set style.display.none;)
  //document.getElementById('show_option').style= 'text-align:center';   // bug?  
  refMsg(); 
	setIns();  
}


// set instructions text depending on user selected task options
function setIns()  
{
  //document.getElementById('instructions').style.display= 'block';		 
	document.getElementById('ins_cross').style.display = isCross ? 'block' : 'none';	
	document.getElementById('ins_demo').style.display= document.getElementById('demo').checked ? 'block' : 'none';	
}	


// formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	var today = new Date();
	var dd = today.getDate();
	var mm = today.getMonth()+1; 
	var yyyy = today.getFullYear();
	if(dd<10) { dd='0'+dd; } 
	if(mm<10) {	mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}


/* Store the results of the last completed task in the user's browser
   set user data into window.localStorage. Parameters:	 
	 points = sum of points in task
	 s_pos => sum of points group by screen position (use only if posit != 1) */  	 
function setStorage(points, s_pos)   
{  
  var endTime = new Date();   // see:  https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Date/getTime
  var task_time= ( endTime.getTime() - startTime.getTime() ) / 1000;  // seconds (convert time difference from milliseconds to seconds)
	task_time= stringTime(task_time);

	var oldItems = JSON.parse(localStorage.getItem('wm4-v2.5')) || []; 
  var id= oldItems.length + 1;     	
	// set new 'record'
  var newItem = {
      'id' : id,
      'date' : setDate(),   //new Date(), 					
			'points' : points,			
			'items' : itemCnt,									
			'demo': document.getElementById('demo').checked ? 1 : 0,			
			'start-level': lc,
      'final-level': hc, 
			'within-level': w,			
			'before' : wait * clock,  
			'display': time * clock, 
      'delay' : delay * clock, 						
			'shape' : document.getElementById('shape').value,			
			'position': posit,
			'monochrome' : monochrome ? 1 : 0,						
      'cross': isCross ? 1 : 0,  
      'task-time': task_time,			
			'print-screen': err,
			'left'  : (posit == 2 && s_pos) ? s_pos[1] : '',
 			'full'  : (posit == 2 && s_pos) ? s_pos[2] : '',
			'right' : (posit == 2 && s_pos) ? s_pos[3] : '',			
			'top-left'   : (posit == 3 && s_pos) ? s_pos[1] : '',
 			'top-right'  : (posit == 3 && s_pos) ? s_pos[2] : '',
			'down-right' : (posit == 3 && s_pos) ? s_pos[3] : '',
			'down-left'  : (posit == 3 && s_pos) ? s_pos[4] : ''						
  };

  oldItems.push(newItem);

  localStorage.setItem('wm4-v2.5', JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems); 
 
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  var col = [];
  for (var i = 0; i < my_json.length; i++) {
    for (var key in my_json[i]) {
        if (col.indexOf(key) === -1) {
            col.push(key);
        }
    }
  } 
  var table = document.createElement('table');
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  var tr = table.insertRow(-1);                   // TABLE ROW.
  for (var i = 0; i < col.length; i++) {
      var th = document.createElement('th');      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (var i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (var j = 0; j < col.length; j++) {
      var tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= '';
  elem.appendChild(table);
}


// clear web browser storage data for this application
function clrStorage()  
{  
  var cnt= localStorage.getItem('wm4-v2.5') ? JSON.parse(localStorage.getItem('wm4-v2.5')).length : 0;                                                        
  if (cnt === 0)  {
    alert('You have no data to delete');
    return;
  }  
  if ( !window.confirm('You have a total of ' + cnt  + ' results stored in your web browser. Continue?') ) { 
   return;
 }                                             
 if (window.confirm('Delete total saved application data:  '  + cnt + '. This action cannot be undone. Final confirmation?') ) 
 {  
   window.localStorage.removeItem('wm4-v2.5'); 
   document.getElementById('store_cnt').textContent= 'All data has been deleted. No data.';
	 document.getElementById('store_oper').style.display= 'none';
 }  
}


// copy data from user web browser to the Clipboard 
function copyStorage()  
{  
  var data= localStorage.getItem('wm4-v2.5');
	if (!data)  {
    alert('You have no data to Copy');
    return;
  }  
	data = JSON.parse(data); 
	var arr= jsonToArray(data);	
	copy2DToClipboard(arr);
}	


// convert json object into 2D array
// data = JSON object
// return 2D array
function jsonToArray(data)  
{
   // extract column (fields) names  
  var total= data.length, col = [], row, res=[];  // res= result (2d array) 
  for (var i = 0; i < total; i++) {
    for (var key in data[i]) {
			if (col.indexOf(key) === -1)  {
				col.push(key);  // col= columns (fields) names
			}
    }
  } 
	res.push(col);  // add field names as 1 row
  var c = col.length;    // column count  
  // add JSON data to the 2d array as rows.
  for (var i = 0; i < total; i++) 
  {
		row= [];
    for (var j = 0; j < c; j++)  {     
      row.push(data[i][col[j]]);
    }
		res.push(row);
  }
	return res;	
}


// https://stackoverflow.com/questions/60698115/how-to-copy-a-javascript-2d-array-to-clipboard-to-paste-it-in-excel
// copy 2D array to Clipboard
function copy2DToClipboard(arr) {
  var csv = '', row, cell;
  for (row = 0; row < arr.length; row++) {
    for (cell = 0; cell < arr[row].length; cell++) {
      csv += (arr[row][cell]+'').replace(/[\n\t]+/g, ' ');
      if (cell+1 < arr[row].length) csv += '\t';
    }
    if (row+1 < arr.length) { 
			csv += '\n'; 
		}
  }
  copyTextToClipboard(csv);
}


// copy 2D array (or text) to Clipboard
function copyTextToClipboard(txt) 
{	
	if (!navigator.clipboard) {
    fallbackCopy(txt);
    return;
  }
  navigator.clipboard.writeText(txt).then(function() {
   //alert('Async: Copying to Clipboard was successful!');
  }, function(err) {
    alert('Oops, could not copy data. Error: ' + err);
  });
}


// copied from https://stackoverflow.com/questions/400212/how-do-i-copy-to-the-clipboard-in-javascript
// fallback copy text to clipboard
function fallbackCopy(txt) 
{
  var ta = document.createElement('textarea');
  ta.value = txt; 
  // Avoid scrolling to bottom
  ta.style.top = '0';
  ta.style.left = '0';
  ta.style.position = 'fixed';
  document.body.appendChild(ta);
  ta.focus();
  ta.select();
  try {
    var successful = document.execCommand('copy');
    var msg = successful ? 'successful' : 'unsuccessful';
  } catch (err) {
    alert('Oops, unable to copy. Error: ' + err);
  }
  document.body.removeChild(ta);
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays user saved results (Window.localStorage application data) in new empty web page (tab) 
function newWin()
{ 
	var data = JSON.parse(localStorage.getItem('wm4-v2.5'))
	createTableJSON(document.getElementById('table_grp'), data);
	var stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	var tbl= document.getElementById('table_grp').innerHTML;	
	var win = window.open('');
  win.document.write('<html><head><title>Saved cumulative data</title>' + stl + '</head><body>' + tbl + '</body></html>');
  win.focus();
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays Latest research results in new empty web page (tab) 
// parameter: id => DOM/html object id (div or table)
function newWinLast()
{ 
  var ind = document.getElementById('last_data_sel').selectedIndex, e;
  if (ind == 0 || posit == 1) {
		e = document.getElementById('table1');  // note: id='table1' is html div element
	}		
	else {
		e = document.getElementById('table2');  // note: id='table2' is html div element	
	}		
	var stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	var tbl= e.innerHTML; 	
	var win = window.open('');
	win.document.write('<html><head><title>Latest research data</title>' + stl + '</head><body>' + tbl + '</body></html>');
	win.focus();
}


// copy only latest task results (data) to the Clipboard 
function copyLast()  
{     
  var ind = document.getElementById('last_data_sel').selectedIndex, tbl;
  if (ind == 0 || posit == 1) {
		tbl = document.getElementById('table1').getElementsByTagName('table')[0];  // note: id='table1' is html div element
	}		
	else {
		tbl = document.getElementById('table2').getElementsByTagName('table')[0];  // note: id='table2' is html div element	
	}				
	if (tbl)  {
		var data= table2DArray(tbl);
		if (data.length > 0) {
			copy2DToClipboard(data);
		}
	}
}	


// https://stackoverflow.com/questions/3065342/how-do-i-iterate-through-table-rows-and-cells-in-javascript
// input argument tbl == html/DOM  table  object
// return 2D array
function table2DArray(tbl) 
{
	var res = [];	 
	for (var i = 0, row; row = tbl.rows[i]; i++) 
	{   //iterate through rows. rows would be accessed using the 'row' variable assigned in the for loop
	  res[i]= [];		
		for (var j = 0, col; col = row.cells[j]; j++) {  //iterate through columns. columns would be accessed using the 'col' variable assigned in the for loop			 
		  res[i][j]= col.textContent;
		}  
	}
	return res;				
}
